#!/bin/bash

cfg_build_configs() {
  declare -a build_configs=($1)
  local OPTS=(--no-validate --no-action-failure --no-check-origin)

  log_info "Generating default configuration files ${build_configs[*]}"
  if ngcpcfg "${OPTS[@]}" build "${build_configs[@]}" ; then
    log_info "OK generating default configuration"
  else
    if ! "${MINIMAL}" ; then
      log_die "Error running 'ngcpcfg ${OPTS[*]} build'"
      exit 1
    else
      log_warn "Errors running 'ngcpcfg ${OPTS[*]} build'"
    fi
  fi
}


cfg_schema_initialise() {
  CONFIG_DIR="$(mktemp -d -t ngcp-installer-cfgschema-XXXXXXXXXX)"
  export CONFIG_DIR

  local cfg_base='/usr/share/ngcp-cfg-schema'
  for file in config constants maintenance network; do
    local file_name="${file}"
    if "${CARRIER}" && [[ "${file}" == 'network' ]]; then
      file_name="${file}_carrier"
    fi
    log_info "Copying '${cfg_base}/${file_name}.yml' to '${CONFIG_DIR}'"
    cp "${cfg_base}/${file_name}.yml" "${CONFIG_DIR}/${file}.yml"
    log_info "Securing ${file}.yml to allow read for root user only"
    chmod 0600 "${CONFIG_DIR}/${file}.yml"
  done

  log_info "Initialization cfg_schema.db"
  sqlite3 "${CONFIG_DIR}/cfg_schema.db" < \
    "${cfg_base}/cfg_scripts/init/init_cfg_schema.sql"
  sqlite3 "${CONFIG_DIR}/cfg_schema.db" < \
    "${cfg_base}/cfg_schema.sql"
}

cfg_apply_cfg_schema() {
  log_info "Executing ngcp-update-cfg-schema"
  ngcp-update-cfg-schema -f -d "${CONFIG_DIR}"
}

# exclude files from etckeeper
cfg_exclude_etc_files() {
  for ign in "${ETCKEEPER_IGNORE[@]}" ; do
    if [ -e "$ign" ] ; then
      log_info "Removed $ign from etckeeper repo"
      (
        set -e
        cd /etc
        git rm --cached "$ign" || true
      )
    fi
    if ! grep -q "${ign##/etc/}" /etc/.gitignore ; then
      echo "${ign##/etc/}" >> /etc/.gitignore
      log_info "Added '${ign##/etc/}' to /etc/.gitignore"
    fi
  done
  (
    set -e
    cd /etc
    git add .gitignore
    if cfg_is_git_clean ".gitignore"; then
      log_info "Git repo /etc is clean"
    else
      log_info "Committing .gitignore file"
      git commit -m 'add ETCKEEPER_IGNORE files to .gitignore' >/dev/null
    fi
  )
}

cfg_ngcpcfg_initialise() {
  log_info "Initializing the configuration framework..."
  if ngcpcfg initialise --without-shared ; then
    log_info "OK initializing the configuration framework"
  else
    log_die "Error running 'ngcpcfg initialise'"
    exit 1
  fi
}

cfg_cloud_tweaks() {
  log_info "Checking for cloud environment"

  if check_running_on_aws ; then
    log_info "* System identified as AWS EC2 instance"
  elif check_running_on_azure ; then
    log_info "* System identified as Microsoft Azure instance"
  else
    log_info "* System doesn't look like a supported cloud instance, skipping cfg_cloud_tweaks"
    return 0
  fi

  log_info "Applying cloud specific changes"

  declare -a pkgs_installed
  pkgs_installed+=(augeas-tools)
  pkgs_installed+=(busybox)
  pkgs_installed+=(cloud-init)
  pkgs_installed+=(cloud-initramfs-growroot)
  pkgs_installed+=(facter)
  pkgs_installed+=(ifupdown)
  pkgs_installed+=(netplan.io)
  pkgs_installed+=(ngcp-linux-cloud-amd64)
  log_info "Making sure we have essential packages installed:"
  if env "${APT_ENV[@]}" apt-get "${APT_PARAMS[@]}" install "${pkgs_installed[@]}" ; then
    log_info "OK"
  else
    log_die "Error installing packages."
    exit 1
  fi

  # we use ngcp-linux-cloud-amd64, so we want to end up with linux-image-cloud-amd64
  # but avoid linux-image-amd64
  log_info "Removing ngcp-linux-amd64 package"
  if env "${APT_ENV[@]}" apt-get "${APT_PARAMS[@]}" purge ngcp-linux-amd64 ; then
    log_info "OK"
  else
    log_die "Error removing ngcp-linux-amd64 package."
    exit 1
  fi

  declare -a pkgs_purge
  pkgs_purge+=(unattended-upgrades)
  log_info "Making sure we do not have certain packages installed:"
  if env "${APT_ENV[@]}" apt-get "${APT_PARAMS[@]}" remove --purge "${pkgs_purge[@]}" ; then
    log_info "OK"
  else
    log_die "Error purging packages."
    exit 1
  fi

  log_info "Adding cloud-init services to ngcpcfg's systemd.custom_preset"
  declare -a cloud_services
  cloud_services+=(cloud-config.service)
  cloud_services+=(cloud-final.service)
  cloud_services+=(cloud-init-local.service)
  cloud_services+=(cloud-init.service)
  for cloud_service in "${cloud_services[@]}" ; do
    ngcpcfg set "${CONFIG_YML}" "systemd.custom_preset.APPEND='${cloud_service}'"
  done
  log_info "OK"

  log_info "Enabling sshd to listen on all IPv4 + IPv6 ports via ngcpcfg's sshd.listen_addresses"
  ngcpcfg set "${CONFIG_YML}" 'sshd.listen_addresses=["0.0.0.0", "::"]'
  log_info "OK"

  if [ -n "$(augtool 'match /files/etc/fstab/*[file="/"]/opt noatime')" ] ; then
    log_info "Mount option 'noatime' already in usage for rootfs, nothing to do."
  else
    log_info "Enabling 'noatime' mount option for rootfs: "
    cat << EOT | augtool >/dev/null
insert opt after /files/etc/fstab/*[file="/"]/opt[last()]
set /files/etc/fstab/*[file="/"]/opt[last()] noatime
save
EOT
    log_info "OK"
  fi

  log_info "Checking for public IPv4 address using facter."
  local PUBLIC_IP
  PUBLIC_IP=$(facter ec2_metadata.public-ipv4 2>/dev/null)

  # facter doesn't provide the public IP on azure, query web API instead
  if [ -z "${PUBLIC_IP:-}" ] ; then
    log_info "Couldn't identify public IPv4 address with facter, trying to identify via API"
    PUBLIC_IP=$(wget --timeout=3 --wait=3 --tries=1 --quiet --no-proxy --header="Metadata:true" -O- \
                "http://169.254.169.254/metadata/instance?api-version=2021-02-01" | \
                jq -r  '.network.interface[0].ipv4.ipAddress[0].publicIpAddress' 2>/dev/null)
  fi

  log_info "Checking for primary network device using facter."
  local PRIMARY_NETDEVICE
  PRIMARY_NETDEVICE=$(facter networking.primary 2>/dev/null)

  if [ -z "${PRIMARY_NETDEVICE}" ] ; then
    echo "Couldn't identify primary network device with facter, falling back to ip route information"
    PRIMARY_NETDEVICE=$(ip route | grep '^default' | sed 's/.* dev //')
  fi

  if ! is_valid_ipv4 "${PUBLIC_IP}" ; then
    log_warn "Could not retrieve IPv4 address, not running ngcp-network to adjust advertised IP address."
  else
    log_info "Setting ${PUBLIC_IP} as advertised IP address on ${PRIMARY_NETDEVICE}."
    ngcp-network --set-interface="${PRIMARY_NETDEVICE}" --advertised-ip="${PUBLIC_IP}"
  fi

  log_info "Generating /etc/ngcp-config/templates/etc/network/interfaces.customtt.tt2"
  tee > /etc/ngcp-config/templates/etc/network/interfaces.customtt.tt2 << EOF
# Generated by ngcpcfg via customtt template for usage on cloud system

# Include files from /etc/network/interfaces.d:
source-directory /etc/network/interfaces.d

# Cloud images dynamically generate config fragments for newly
# attached interfaces. See /etc/udev/rules.d/75-cloud-ifupdown.rules
# and /etc/network/cloud-ifupdown-helper. Dynamically generated
# configuration fragments are stored in /run:
source-directory /run/network/interfaces.d

EOF

  if ! [ -f /etc/network/cloud-ifupdown-helper ] ; then
    log_info "No /etc/network/cloud-ifupdown-helper present, setting up the file"
    tee > /etc/network/cloud-ifupdown-helper << "EOF"
#!/bin/bash
# NOTE: deployed via NGCP installer
# Copyright 2019 Noah Meyerhans <noahm@debian.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

template="/etc/network/cloud-interfaces-template"
cfgdir="/run/network/interfaces.d"
mkdir -p "$cfgdir"

log() {
    logger --tag "cloud-ifupdown-helper" --priority daemon.debug "$@"
}

do_setup() {
    local working="$cfgdir/.$INTERFACE"
    local final="$cfgdir/$INTERFACE"

    if ifup --no-act "$INTERFACE" > /dev/null 2>&1; then
        # interface is already known to ifupdown, no need to generate cfg
        log "Skipping configuration generation for $INTERFACE"
        return 0
    fi

    sed "s/\\\$INTERFACE/$INTERFACE/g" "$template" > "$working"
    mv "$working" "$final"
    log "Generated configuration for $INTERFACE"

    # Raise the interface here, because we can't rely on the RUN
    # entries from /lib/udev/rules.d/80-ifupdown.rules running after
    # this script:
    systemctl --no-block start $(systemd-escape --template ifup@.service $INTERFACE)
}

if [ -z "$INTERFACE" ]; then
    log 'Bad ifupdown udev helper invocation: $INTERFACE is not set'
    exit 1
fi

do_setup

exit 0
EOF
    chmod 755 /etc/network/cloud-ifupdown-helper
  fi

  if ! [ -f /etc/network/cloud-interfaces-template ] ; then
    log_info "No /etc/network/cloud-interfaces-template present, setting up the file"
    tee > /etc/network/cloud-interfaces-template << "EOF"
# NOTE: deployed via NGCP installer based on /etc/network/cloud-interfaces-template
auto $INTERFACE
allow-hotplug $INTERFACE

iface $INTERFACE inet dhcp
EOF
  fi

  if ! [ -f /etc/udev/rules.d/75-cloud-ifupdown.rules  ] ; then
    log_info "No /etc/udev/rules.d/75-cloud-ifupdown.rules present, setting up the file"
    tee > /etc/udev/rules.d/75-cloud-ifupdown.rules  << "EOF"
# NOTE: deployed via NGCP installer
# Handle allow-hotplug interfaces
SUBSYSTEM=="net", ACTION=="add", ENV{INTERFACE}=="neth*|eth*|en*", RUN+="/etc/network/cloud-ifupdown-helper"
EOF
  fi

  log_info "Disabling systemd-networkd services"
  systemctl disable systemd-networkd.socket
  systemctl disable systemd-networkd.service
  systemctl mask    systemd-networkd.service
}

cfg_physical_server_tweaks() {
  if check_running_inside_virt_env ; then
    log_info "System doesn't seem to be physical system, skipping physical server tweaks"
  else
    log_info "Physical system detected, enabling smartmontools monitoring"
    ngcpcfg set "${CONFIG_YML}" "monitoring.smartmontools.enable=yes"
  fi
}

cfg_build_templates() {
  cfg_build_configs

  log_info "Saving config files..."
  cfg_exclude_etc_files
  pushd /etc &>/dev/null
  if cfg_is_git_clean; then
    log_info "/etc repo is clean"
  else
    etckeeper commit "Initial config"
  fi
  popd &>/dev/null
  ngcpcfg --no-db-sync commit "Config files were created"
  log_info "Done saving config files"
}

cfg_generate_certs() {
  if "${MINIMAL}" ; then
    log_info "minimal option detected, skip step"
    return 0
  fi

  log_info "Creating group ssl-cert and add users into it"
  mkdir -p "$CERT_PATH"
  find "$CERT_PATH" -type d -exec chmod 0750 {} \;
  chown root:ssl-cert "$CERT_PATH"
  for u in $CERT_USERS ; do
    if ! groups "$u"| grep -q ssl-cert 2>/dev/null ; then
      log_info "Adding user $u to ssl-cert"
      adduser "$u" ssl-cert
    fi
  done

  log_info "Generating openssl keys"
  if [ ! -x /usr/share/ngcp-panel-tools/generate_ssl_keys.sh ]; then
    log_error "No /usr/share/ngcp-panel-tools/generate_ssl_keys.sh found"
    return 1
  fi

  if [ -f "${CERT_PATH}/myserver.crt" ] && \
     [ -f "${CERT_PATH}/myserver.csr" ] && \
     [ -f "${CERT_PATH}/myserver.key" ]
  then
    log_info "Self-signed WEB keys 'myserver' already exist, continue..."
  else
    log_info "Generating self-signed WEB keys 'myserver':"
    if /usr/share/ngcp-panel-tools/generate_ssl_keys.sh "${CERT_PATH}" /usr/share/ngcp-panel-tools myserver ; then
      log_info "OK generating self-signed 'myserver' keys"
    else
      log_error "Problem generating self-signed 'myserver' keys"
      return 1
    fi
  fi

  if [ -f "${CERT_PATH}/api_ca.crt" ] && \
     [ -f "${CERT_PATH}/api_ca.key" ]
  then
    log_info "REST auth keys 'api_ca' already exist, continue..."
  else
    log_info "Generating REST auth keys 'api_ca':"
    if SKIP_CSR=true /usr/share/ngcp-panel-tools/generate_ssl_keys.sh "${CERT_PATH}" ; then
      chmod a+r "${CERT_PATH}"/api_ca.*
      log_info "OK generating REST auth keys 'api_ca'"
    else
      log_error "Problem generating REST auth keys 'api_ca'"
      return 1
    fi
  fi

  if [ -f "${CERT_PATH}/dhparam.pem" ] ; then
    log_info "dhparams.pem already exists, continue..."
  else
    log_info "Generating dhparams.pem (this is going to take a long time)"
    openssl dhparam -dsaparam -out "${CERT_PATH}/dhparam.pem" 4096
    chown root:ssl-cert "${CERT_PATH}/dhparam.pem"
    chmod 640 "${CERT_PATH}/dhparam.pem"
    log_info "Done generating dhparams.pem"
  fi

  cfg_save_defaults "ssl certs" "--no-db-sync"
}

cfg_save_defaults() {
  local message=${1:-}
  declare -a opts=(${2:-})

  if [ "$(ngcpcfg diff 2>/dev/null | wc -l)" != "0" ] ; then
    log_info "Committing config changes for $message"
    ngcpcfg "${opts[@]}" commit "changes for $message"
    etckeeper commit "changes for $message" || true
    log_info "Done"
  fi
}

cfg_systemctl_preset_all() {
  if "${DOCKER_BUILD}"; then
    log_info "It is Docker build, no need to configure preset"
    return 0
  fi
  log_info "Removing /etc/systemd/system-preset/01-installer.preset"
  rm -f /etc/systemd/system-preset/01-installer.preset
  log_info "Calling systemd preset to configure services on the next boot:"
  systemctl preset-all
}

cfg_set_network_config_ce() {
  log_info "Update ${NETWORK_YML} with real CE network parameters"

  if [[ "${USE_EXISTING_YML}" =~ network ]]; then
    log_info "It is specified to use prepared network.yml, do not change in ${FUNCNAME[0]}"
    return
  fi

  ngcp-network --set-interface="${EXTERNAL_DEV}" \
    --ip="${EADDR}" --netmask="${EXTERNAL_NETMASK}" --gateway="${GW}" --hwaddr=auto

  for entry in ${NAMESERVER} ; do
    ngcp-network --set-interface="${EXTERNAL_DEV}" --dns="${entry}"
  done

  vm_net_configuration 'spce'
}

cfg_build_net_configs() {
  log_info "Generating default configuration files /etc/network/interfaces /etc/hosts /etc/iproute2/rt_tables"
  cfg_build_configs "/etc/network/interfaces /etc/hosts /etc/iproute2/rt_tables"
}

cfg_apply_configs_docker() {
  log_info "It is Docker build, can't apply configuration due to systemctl limitation"

  log_info "Building configuration"
  if ! ngcpcfg build ; then
    log_die "Can not build configuration"
    exit 1
  fi

  log_info "Committing configuration"
  declare -a ngcpcfg_opts
  if "${DOCKER_PURE_BUILD}"; then
    ngcpcfg_opts+=(--no-db-sync)
  fi
  if ! ngcpcfg "${ngcpcfg_opts[@]}" commit "Installer building/committing configuration" ; then
    log_die "Can not commit configuration"
    exit 1
  fi
}

cfg_apply_configs_default() {
  log_info "Applying configuration"
  if ! ngcpcfg apply --force-all-services "Installer applying configuration" ; then
    log_die "Can not apply configuration"
    exit 1
  fi
}

cfg_apply_configs() {
  if "${DOCKER_BUILD}"; then
    cfg_apply_configs_docker
  else
    cfg_apply_configs_default
  fi
}

get_underlying_net_dev() {
  local device
  device="${1}"
  if [[ -z "${device}" ]]; then
    log_die "Interface name has to be passed to get_underlying_net_dev"
    exit 1
  fi

  local underlying_device
  underlying_device="$( ip a show "${device}" | sed -r 's/.+vlan[0-9]+@(.+):.+$/\1/;t;d' )"
  if [[ -z "${underlying_device}" ]]; then
    log_die "The underlying device was not detected"
    exit 1
  fi
  echo "${underlying_device}"

  return 0
}

cfg_normalize_net_dev() {
  if [[ "${EXTERNAL_DEV}" =~ ^vlan ]]; then
    EXTERNAL_DEV="$( get_underlying_net_dev "${EXTERNAL_DEV}" )"
  fi
}

cfg_common_configuration() {
  for tuning_arg in "${USE_TUNING[@]}"; do
    if [[ "${tuning_arg}" == 'low-perf-profile' ]]; then
      log_info "Enabling low performance mode as specified in --use-tuning"
      ngcp-toggle-performance-config --low-performance --skip-commit
    fi
  done

  if [[ "${USE_EXISTING_YML}" =~ config ]]; then
    log_info "It is specified to use prepared config.yml, skipping further changes in ${FUNCNAME[0]}"
    return
  fi

  log_info "Change default value of openvpn.keys_filename_template"
  ngcpcfg set "${CONFIG_YML}" "openvpn.keys_filename_template='/etc/ngcp-config/shared-files/openvpn/client/customer-%h_scep.%t'"

  log_info "Save size of 'rootfs', 'fallbackfs' and 'swapfile' partitions to ${CONFIG_YML}"
  ngcpcfg set "${CONFIG_YML}" \
    "bootenv.netscript.fallbackfssize=${FALLBACKFS_SIZE}" \
    "bootenv.netscript.rootfssize=${ROOTFS_SIZE}" \
    "bootenv.netscript.swapfilesize=${SWAPFILE_SIZE_MB}"

  local POOL_SIZE_MB
  if ${PRO} || ${CARRIER}; then
    log_info "Preparing dynamic configuration settings for MySQL:"
    local TOTAL_MEM
    TOTAL_MEM=$(awk '/^MemTotal:/ {print $2}' /proc/meminfo)

    log_info "    Total memory detected: ${TOTAL_MEM}"
    [ -z "${INNODB_BUFFER_POOL_FACTOR}" ] && INNODB_BUFFER_POOL_FACTOR="0.5"

    log_info "    Innodb poolsize factor set to: ${INNODB_BUFFER_POOL_FACTOR}"
    POOL_SIZE_MB=$(echo | awk -v tm="${TOTAL_MEM}" -v mf=${INNODB_BUFFER_POOL_FACTOR} '{print int(tm*mf/1000)}')
    POOL_SIZE_MB="${POOL_SIZE_MB}M"
  else
    # in CE systems we hardcode the innodb_pool_size setting
    log_info "CE system detected, setting static configuration for MySQL innodb_pool_size:"
    POOL_SIZE_MB="64M"
  fi
  log_info "Save innodb_pool_size ${POOL_SIZE_MB} to ${CONFIG_YML}"
  ngcpcfg set "${CONFIG_YML}" "database.bufferpoolsize=${POOL_SIZE_MB}"

  # we need to decrease default high performance config
  # for low performance virtual environment (e.g. Vagrant)
  if "${ADJUST_FOR_LOW_PERFORMANCE}" ; then
    log_info "Enabling low performance mode"
    /usr/sbin/ngcp-toggle-performance-config --low-performance --skip-commit
  fi
  if "${ENABLE_VM_SERVICES}" ; then
    # we do enable some services on Vagrant VMs for testing
    enable_vm_services
  fi

  # enable qemu-gest-agent integration
  vm_qemu_guest_agent

  if [[ -n "${NGCP_PPA}" ]]; then
    local old_IFS="${IFS}"
    IFS=','
    declare -a ppas=( ${NGCP_PPA} )
    for ppa in "${ppas[@]}"; do
      ngcpcfg set "${CONFIG_YML}" bootenv.ppa.APPEND="{'name','${ppa}'}"
    done
    IFS="${old_IFS}"
  fi
}

cfg_is_git_clean() {
  log_debug "call 'git status --porcelain=v2 $*' in $(pwd). it must have no output"
  if [[ -z "$(git status --porcelain=v2 "$@")" ]]; then
    return 0
  else
    return 1
  fi
}

cfg_place_yml_files() {
  if "${MINIMAL}" ; then
    mkdir -p "${CONFIGPATH}"
  fi

  log_info "Copying '${CONFIG_DIR}/cfg_schema.db' to '${CFG_SCHEMA_FILE}'"
  cp "${CONFIG_DIR}/cfg_schema.db" "${CFG_SCHEMA_FILE}"

  if [[ "${USE_EXISTING_YML}" =~ config ]]; then
    log_info "USE_EXISTING_YML config is specified so config.yml is not copied"
  else
    log_info "Copying '${CONFIG_DIR}/config.yml' to '${CONFIG_YML}'"
    cp "${CONFIG_DIR}/config.yml" "${CONFIG_YML}"
  fi

  if [[ "${USE_EXISTING_YML}" =~ constants ]]; then
    log_info "USE_EXISTING_YML constants is specified so constants.yml is not copied"
  else
    log_info "Copying '${CONFIG_DIR}/constants.yml' to '${CONSTANTS_YML}'"
    cp "${CONFIG_DIR}/constants.yml" "${CONSTANTS_YML}"
  fi

  if [[ "${USE_EXISTING_YML}" =~ maintenance ]]; then
    log_info "USE_EXISTING_YML maintenance is specified so maintenance.yml is not copied"
  else
    log_info "Copying '${CONFIG_DIR}/maintenance.yml' to '${MAINTENANCE_YML}'"
    cp "${CONFIG_DIR}/maintenance.yml" "${MAINTENANCE_YML}"
  fi

  if [[ "${USE_EXISTING_YML}" =~ network ]]; then
    log_info "USE_EXISTING_YML network is specified so network.yml is not copied"
  else
    log_info "Copying '${CONFIG_DIR}/network.yml' to '${NETWORK_YML}'"
    cp "${CONFIG_DIR}/network.yml" "${NETWORK_YML}"
  fi
}

cfg_generate_ngcp_panel_keys() {
  if which ngcp-panel-create-keys ; then
    log_info "Creating asymmetric keys for ngcp-panel"
    ngcp-panel-create-keys
    cfg_save_defaults "ngcp-panel keys" "--no-db-sync"
  else
    if "${MINIMAL}" ; then
      log_info "Skiping creation of asymmetric keys for ngcp-panel, ngcp-panel not installed"
    else
      log_error "ngcp-panel-create-keys not found"
      exit 1
    fi
  fi
}

cfg_validate_arguments() {
  if "${JOIN_CLUSTER}" && [[ -n "${USE_EXISTING_YML}" ]]; then
    log_error "Option --join and --use-existing-yml= both enabled but they are mutually exclusive"
    exit 1
  fi
  for tuning_arg in "${USE_TUNING[@]}"; do
    case "${tuning_arg}" in
      low-perf-profile)
        :
      ;;
      *)
        log_error "Unsupported tuning option ${tuning_arg}, exiting..."
        exit 1
      ;;
    esac
  done
}
